use core::fmt::Write;
use core::str::from_utf8;
use embassy_net::tcp::TcpSocket;
use embassy_time::{Duration, Timer};
use embedded_io_async::Write as _;
use heapless::Vec;
use log::{info, warn};

use crate::apps::{App, chat, index::IndexApp, ttt};

#[embassy_executor::task(pool_size = 2)]
pub async fn ttt_listen_task(stack: embassy_net::Stack<'static>, team: ttt::Team, port: u16) {
    listen_task(stack, ttt::TttApp::new(team), port).await
}

#[embassy_executor::task(pool_size = 2)]
pub async fn index_listen_task(stack: embassy_net::Stack<'static>, port: u16) {
    listen_task(stack, IndexApp, port).await
}

#[embassy_executor::task(pool_size = 1)]
pub async fn chat_listen_task(stack: embassy_net::Stack<'static>, port: u16) {
    listen_task(stack, chat::ChatApp::new(), port).await
}

pub async fn listen_task(stack: embassy_net::Stack<'static>, mut app: impl App, port: u16) {
    // loop {
    //     info!("team:{:?}", team);
    //     Timer::after_millis(0).await;
    // }
    let mut rx_buffer = [0; 8192];
    let mut tx_buffer = [0; 8192];
    let mut buf = [0; 1024];
    let mut res_head_buf = Vec::<u8, 1024>::new();

    loop {
        Timer::after_secs(0).await;
        let mut socket = TcpSocket::new(stack, &mut rx_buffer, &mut tx_buffer);
        socket.set_timeout(Some(Duration::from_secs(30)));

        info!("Socket {}: Listening on TCP:{}...", app.socket_name(), port);
        if let Err(e) = socket.accept(port).await {
            warn!("accept error: {:?}", e);
            continue;
        }

        info!(
            "Socket {}: Received connection from {:?}",
            app.socket_name(),
            socket.remote_endpoint()
        );

        loop {
            Timer::after_secs(0).await;
            let n = match socket.read(&mut buf).await {
                Ok(0) => {
                    warn!("read EOF");
                    break;
                }
                Ok(n) => n,
                Err(e) => {
                    warn!("Socket {}: read error: {:?}", app.socket_name(), e);
                    break;
                }
            };

            let (headers, content) = match from_utf8(&buf[..n]) {
                Ok(b) => match b.split_once("\r\n\r\n") {
                    Some(t) => t,
                    None => (b, ""),
                },
                Err(_) => {
                    warn!("Non utf8 http request");
                    break;
                }
            };

            // let mut headers = headers.split(|x| *x == b'\n');
            let (request_type, path) = match headers.lines().next() {
                None => {
                    warn!("Empty request");
                    break;
                }
                Some(l1) => {
                    let mut l1 = l1.split(' ');
                    (
                        match l1.next() {
                            Some("GET") => HttpRequestType::Get,
                            Some("POST") => HttpRequestType::Post,
                            Some(t) => {
                                warn!("Unknown request type : {}", t);
                                break;
                            }
                            None => {
                                warn!("No request type");
                                break;
                            }
                        },
                        match l1.next() {
                            Some(path) => path,
                            None => {
                                warn!("No path");
                                break;
                            }
                        },
                    )
                }
            };

            info!(
                "Socket {}: {:?} request for {}",
                app.socket_name(),
                request_type,
                path
            );
            Timer::after_secs(0).await;

            let (code, res_type, res_content): (HttpResCode, &str, &[u8]) = match path {
                "/htmx.js" => (
                    HttpResCode::Ok,
                    "javascript",
                    #[cfg(debug_assertions)]
                    include_bytes!("../static/htmx.js"),
                    #[cfg(not(debug_assertions))]
                    include_bytes!("../static/htmx.min.js"),
                ),
                p => app.handle_request(p, request_type, content).await,
            };

            res_head_buf.clear();
            if let Err(e) = write!(
                &mut res_head_buf,
                "{}\r\n\
                Content-Type: text/{}\r\n\
                Content-Length: {}\r\n\r\n",
                Into::<&str>::into(code),
                res_type,
                res_content.len()
            ) {
                warn!("res buffer write error: {:?}", e);
                break;
            }

            match socket.write_all(&res_head_buf).await {
                Ok(()) => {}
                Err(e) => {
                    warn!("write error: {:?}", e);
                    break;
                }
            };
            match socket.write_all(&res_content).await {
                Ok(()) => {}
                Err(e) => {
                    warn!("write error: {:?}", e);
                    break;
                }
            };
        }
    }
}

#[derive(Clone, Copy, Debug)]
pub enum HttpRequestType {
    Get,
    Post,
}
impl Into<&str> for HttpRequestType {
    fn into(self) -> &'static str {
        match self {
            Self::Get => "GET",
            Self::Post => "POST",
        }
    }
}

#[derive(Debug, Clone, Copy)]
pub enum HttpResCode {
    Ok,
    NoContent,
    BadRequest,
    NotFound,
    Forbidden,
}
impl Into<&str> for HttpResCode {
    fn into(self) -> &'static str {
        match self {
            HttpResCode::Ok => "HTTP/1.1 200 OK",
            HttpResCode::NoContent => "HTTP/1.1 204 NO CONTENT",
            HttpResCode::BadRequest => "HTTP/1.1 400 BAD REQUEST",
            HttpResCode::NotFound => "HTTP/1.1 404 NOT FOUND",
            HttpResCode::Forbidden => "HTTP/1.1 403 FORBIDDEN",
        }
    }
}
